from .tokenizer import TokenizerPipeline
from .tokenizer import CJKProcessor
from .tokenizer import RegexTokenizer
from . import lexicon

wikitext_split_w_cjk = TokenizerPipeline(RegexTokenizer(lexicon.WIKITEXT_SPLIT_LEXICON), # noqa
                                         CJKProcessor(lexicon.CJK_LEXICON))



import re
import jieba as ch_jieba
from sudachipy import tokenizer as jp_tokenizer
from sudachipy import dictionary as jp_dictionary
from konlpy.tag import Okt as ko_okt


CH_JIEBA = ch_jieba.Tokenizer()
JAP_SUDACHY = None
KOR_KONLPY_OKT = None


def get_ch_tokenizer():
    if CH_JIEBA.initialized is False:
        CH_JIEBA.initialize()
    return CH_JIEBA


def get_jap_tokenizer():
    global JAP_SUDACHY
    if JAP_SUDACHY is None:
        JAP_SUDACHY = jp_dictionary.Dictionary().create()
    return JAP_SUDACHY


def get_kor_tokenizer():
    global KOR_KONLPY_OKT
    if KOR_KONLPY_OKT is None:
        KOR_KONLPY_OKT = ko_okt()
    return KOR_KONLPY_OKT


def lng_decision(text, cjk_lexicon, lng_frac_par=0.25, bck_ch_frac=0.5):
    regex_cjk = re.compile(cjk_lexicon['cjk'])
    regex_ch = re.compile(cjk_lexicon['chinese'])
    regex_japanese = re.compile(cjk_lexicon['japanese'])
    regex_korean = re.compile(cjk_lexicon['korean'])

    cjk_symbols = len(regex_cjk.findall(text))

    if cjk_symbols > 0:

        ch_symbols = len(regex_ch.findall(text))
        jap_symbols = len(regex_japanese.findall(text))
        kor_symbols = len(regex_korean.findall(text))
        char_lng_frac = {'japanese': jap_symbols/cjk_symbols,
                         'korean': kor_symbols/cjk_symbols}
        max_char_lng_frac = max(char_lng_frac, key=char_lng_frac.get)
        # check if at least 1/4 of chars are other than chinese,
        # if not and chinese chars are more than 1/2 ->
        # run chinese tokenizer, else -> do nothing
        if char_lng_frac[max_char_lng_frac] > lng_frac_par:
            language = max_char_lng_frac

        elif ch_symbols/cjk_symbols > bck_ch_frac:
            language = 'chinese'

        else:
            language = 'other'
    else:
        language = 'other'

    return language


def CJK_tokenization(text, language):
    if language == 'chinese':
        seg = get_ch_tokenizer()
        processed_text = seg.lcut(text, cut_all=False)

    if language == 'japanese':
        mode = jp_tokenizer.Tokenizer.SplitMode.B
        seg = get_jap_tokenizer()
        processed_text = [m.surface()
                          for m in seg.tokenize(str(text),
                          mode)]

    if language == 'korean':
        seg = get_kor_tokenizer()
        processed_text = [m[0] for m in seg.pos(text)]

    else:
        processed_text = processed_text

    return processed_text



import re
import yamlconf
from .token import Token
from . import cjk_tokenization


class Tokenizer:
    """
    Constructs a tokenizaton strategy.
    """
    def tokenize(self, text, token_class=Token):
        """
        Tokenizes a text.
        """
        raise NotImplementedError()

    @classmethod
    def from_config(cls, config, name, section_key="tokenizers"):
        section = config[section_key][name]
        if 'module' in section:
            return yamlconf.import_module(section['module'])
        else:
            Tokenizer = yamlconf.import_module(section['class'])
            return Tokenizer.from_config(config, name, section_key)


class RegexTokenizer(Tokenizer):
    """
    Uses a lexicon of regular expressions and names to tokenize a text string.
    """
    def __init__(self, lexicon):
        self.lexicon = lexicon
        self.regex = re.compile('|'.join('(?P<{0}>{1})'.format(name, pattern)
                                         for name, pattern in lexicon))

    def tokenize(self, text, token_class=None):
        return [t for t in self._tokenize(text, token_class=token_class)]

    def _tokenize(self, text, token_class=None):
        """
        Tokenizes a text

        :Returns:
            A `list` of tokens
        """
        token_class = token_class or Token
        tokens = {}

        for i, match in enumerate(self.regex.finditer(text)):
            value = match.group(0)

            try:
                token = tokens[value]
            except KeyError:
                type = match.lastgroup
                token = token_class(value, type=type)
                tokens[value] = token

            yield token


class TokenizerPipeline(Tokenizer):
    def __init__(self, tokenizer, *token_processors):
        self.tokenizer = tokenizer  # start of the pipeline
        self.token_processors = token_processors

    def tokenize(self, text):
        tokens = self.tokenizer.tokenize(text)
        for token_processor in self.token_processors:
            tokens = token_processor.process(tokens)
        return tokens


class TokenProcessor:
    def process(self, tokenized_text):
        raise NotImplementedError()


class CJKProcessor(TokenProcessor):
    """
    Uses a cjk_lexicon to decide which tokenizer should be used
    (Chinese, Japanese or Korean).
    """
    def __init__(self, cjk_lexicon, lng_frac_par=0.25, bck_ch_frac=0.5):
        self.cjk_lexicon = cjk_lexicon
        self.lng_frac_par = lng_frac_par
        self.bck_ch_frac = bck_ch_frac

    def process(self, tokenized_text, token_class=None):
        token_class = token_class or Token
        language = self._lng_decision(tokenized_text)
        if language != 'other':
            processed_tokens = self._cjk_processing(tokenized_text,
                                                    language=language,
                                                    token_class=token_class)
        else:
            processed_tokens = tokenized_text
        return processed_tokens

    def _lng_decision(self, tokenized_text):
        text = "".join([tok[:]
                        for tok in tokenized_text
                        if tok.type == 'cjk_word'])
        language = cjk_tokenization.lng_decision(text, self.cjk_lexicon,
                                                 self.lng_frac_par,
                                                 self.bck_ch_frac)
        return language

    def _cjk_processing(self, tokenized_text, language, token_class=None):
        token_class = token_class or Token
        cjk_word_indices = list(filter(
                                lambda x:
                                tokenized_text[x].type == 'cjk_word',
                                range(len(tokenized_text))))

        # go from the last to first,
        # "unpack" CJK "words" and assign "cjk_word" to them
        for i in cjk_word_indices[::-1]:
            proc_cjk_token = cjk_tokenization.CJK_tokenization(
                                           tokenized_text[i], language)
            tokenized_text[i:i+1] = [token_class(word, type="cjk_word")
                                     for word in proc_cjk_token]

        return tokenized_text



#       REGEXP CHEATSHEET
#
# ?     ab? will match either ‘a’ or ‘ab’
# []	Used to indicate a set of characters. In a set
# ^	    If the first character of the set is '^', all the
#       characters that are not in the set will be matched. For example,
#       [^5] will match any character except '5'
# \w	matches any alphanumeric character and the underscore; this is
#       equivalent to the set [a-zA-Z0-9_]
# \W	matches any non-alphanumeric character; this is equivalent to
#       the set [^a-zA-Z0-9_]
# \d	matches any decimal digit; this is equivalent to the set [0-9]
# |	    or
# \s	matches any whitespace character, this is equivalent to the
#       set [ \t\n\r\f\v]
# $	    Matches the end of the string or just before the newline at the
#       end of the string
# *     match 0 or more repetitions of the preceding RE, as many repetitions as
#       are possible. ab* will match ‘a’, ‘ab’, or ‘a’ followed by any number
#       of ‘b’s.
# \b    used to find a match at the beginning or end of a word.

# Unified Ideographs, Unified Ideographs Ext A,
# Unified Ideographs Ext. B, Compatibility Ideographs,
# Compatibility Ideographs Suppl.
chinese_n_misc_char = r'\u4E00-\u62FF' + \
                      r'\u3400-\u4DFF' + \
                      r'\U00020000-\U000215FF' + \
                      r'\uF900-\uFAFF' + \
                      r'\U0002F800-\U0002FA1F' + \
                      r'\u6300-\u77FF' + \
                      r'\u7800-\u8CFF' + \
                      r'\u8D00-\u9FCC' + \
                      r'\U00021600-\U000230FF' + \
                      r'\U00023100-\U000245FF' + \
                      r'\U00024600-\U000260FF' + \
                      r'\U00026100-\U000275FF' + \
                      r'\U00027600-\U000290FF' + \
                      r'\U00029100-\U0002A6DF' + \
                      r'\u4E00-\u9FCB' + \
                      r'\uF900-\uFA6A' + \
                      r'\u3220-\u3243' + \
                      r'\u3280-\u337F'

# Hiragana, Katakana, Kanji, Kanji radicals,
# Katakana and Punctuation (Half Width), Misc Symbols and Chars
jap_char = r'\u3041-\u3096' + \
           r'\u30A0-\u30FF' + \
           r'\u3400-\u4DB5' + \
           r'\u2E80-\u2FD5' + \
           r'\uFF5F-\uFF9F' + \
           r'\u31F0-\u31FF'

# hangul syllables, hangul jamo, hangul, hangul, hangul
kor_char = r'\uAC00-\uD7AF' + \
           r'\u1100-\u11FF' + \
           r'\u3130–\u318F' + \
           r'\uA960–\uA97F' + \
           r'\uD7B0–\uD7FF'

cjk_char = chinese_n_misc_char + jap_char + kor_char

cjk_word = r'(?=[' + cjk_char + r'])' + \
       r'[' + cjk_char + r']*' + \
       r'(?:[\'’](?:[' + cjk_char + r']+|(?=(?:$|\s))))*'

devangari_char = r'\u0901-\u0963'

arabic_char = r'\u0601-\u061A' + \
              r'\u061C-\u0669' + \
              r'\u06D5-\u06EF'

bengali_char = r'\u0980-\u09FF'

combined_char = devangari_char + \
                arabic_char + \
                bengali_char


# (anything that is not(non-alphanmeric, decimal digit) or is combined_char) +
# any number of combined_char repetitions +
# any number of combined_char repetitions encapsulated in apostrophes

word = r'(?:[^\W\d]|[' + combined_char + r'])' + \
       r'[\w' + combined_char + r']*' + \
       r'(?:[\'’](?:[\w' + combined_char + r']+|(?=(?:$|\s))))*'

plain_proto = [r'bitcoin', r'geo', r'magnet', r'mailto', r'news', r'sips?',
               r'tel', r'urn']

slashed_proto = [r'https?', r'ftp', r'ftps', r'git', r'gopher', r'ircs?',
                 r'mms', r'nntp', r'redis', r'sftp', r'ssh', r'svn', r'telnet',
                 r'worldwind', r'xmpp']

address = r'[^\s/$.?#][^|<>\s{}]*'

url = (
    r'(?:' +  # noqa
        r'(?:' + '|'.join(plain_proto) + r')\:|' +  # noqa
        r'(?:(?:' + '|'.join(slashed_proto) + r')\:)?\/\/' +
    r')' + address
)

number = r'[0-9][0-9\.\,]*(e[0-9]+)*|[\.\,][0-9][0-9\.\,]*(e[0-9]+)*'

TEXT_SPLIT_LEXICON = [
    ('word', r'[^\W\d]+'),
    ('number', number),
    ('period', r'\.'),
    ('qmark', r'\?'),
    ('epoint', r'!'),
    ('comma', r','),
    ('colon', r':'),
    ('scolon', r';'),
    ('break', r'(\n|\n\r|\r\n)\s*(\n|\n\r|\r\n)+'),
    ('whitespace', r'[\n\r\s]+'),
    ("etc", r".")
]

CJK_LEXICON = {
    'cjk': r'[' + cjk_char + r']',
    'chinese': r'[' + chinese_n_misc_char + r']',
    'japanese': r'[' + jap_char + r']',
    'korean': r'[' + kor_char + r']',
}

WIKITEXT_SPLIT_LEXICON = [
    ('break', r'(?:\n\r?|\r\n)\s*(?:\n\r?|\r\n)+'),
    ('whitespace', r'(?:\n\r?|[^\S\n\r]+)'),
    ("url", url),
    ("equals", r"=+"),
    ("bar", r"\|"),
    ('entity', r'&[a-z][a-z0-9]*;'),
    ('ref_open', r'<ref\b(?:\/(?!>)|[^>\/])*>'),
    ('ref_close', r'</ref\b[^>]*>'),
    ('ref_singleton', r'<ref\b(?:\/(?!>)|[^/>])*\/>'),
    ('tag', r'</?([a-z][a-z0-9]*)\b[^>]*>'),
    ('number', number),
    ("bold", r"'''"),
    ("italic", r"''"),
    ('japan_punct', r'[\u3000-\u303F]'),
    ('cjk_word', cjk_word),
    ('word', word),
    ('tab_open', r'\{\|'),
    ('tab_close', r'\|\}'),
    ('dbrack_open', r'\[\['),
    ('dbrack_close', r'\]\]'),
    ('brack_open', r'\['),
    ('brack_close', r'\]'),
    ('paren_open', r'\('),
    ('paren_close', r'\)'),
    ('dcurly_open', r'\{\{'),
    ('dcurly_close', r'\}\}'),
    ('curly_open', r'\{'),
    ('curly_close', r'\}'),
    ('period', r'\.+'),
    ('qmark', r'\?+'),
    ('epoint', r'!+'),
    ('comma', r',+'),
    ('colon', r':+'),
    ('scolon', r';+'),
    ('comment_start', r'<!--'),
    ('comment_end', r'-->'),
    ('danda', r'।|॥'),
    ("etc", r"."),
]
